package minischeme

import scala.util.parsing.input.Position
import SymbolicTreeModule._

/**
 * Guarantees that all primitives appearing in the program are always
 * applied, by performing eta-expansion when necessary. For example,
 * the tree:
 *
 *   (fold + 0 lst)
 *
 * is transformed to:
 *
 *   (fold (lambda (x1 x2) (+ x1 x2)) 0 lst)
 *
 * At the same time, Apply nodes that refer to primitives are
 * rewritten to Prim nodes.
 *
 * @author Michel Schinz
 */

object EtaExpander extends (List[Tree] => List[Tree]) {
  def apply(trees: List[Tree]): List[Tree] =
    trees map transform

  private def transform(t: Tree): Tree = copyPos(t) { t match {
    case Define(name, value) =>
      Define(name, transform(value))
    case Lambda(args, body) =>
      Lambda(args, transform(body))
    case Let(name, value, body) =>
      Let(name, transform(value), transform(body))
    case Begin(body) =>
      Begin(body map transform)
    case If(cond, thenp, elsep) =>
      If(transform(cond), transform(thenp), transform(elsep))
    case And(c1, c2) =>
      And(transform(c1), transform(c2))
    case Or(c1, c2) =>
      Or(transform(c1), transform(c2))
    case Apply(Ident(sym), args) if (sym.kind == PrimitiveKind) =>
      Prim(sym, args map transform)
    case Apply(fun, args) =>
      Apply(transform(fun), args map transform)
    case Prim(sym, args) =>
      Prim(sym, args map transform)
    case Ident(sym) if (sym.kind == PrimitiveKind) =>
      etaExpand(sym, t.pos)
    case Ident(_) | Num(_) =>
      t
  } }

  private def etaExpand(sym: Symbol, pos: Position): Tree = {
    val arity = Primitives.arity(sym)
    if (arity < 0)
      Report.error(pos, "primitive " + sym.name + " cannot be eta-expanded!")
    val pSyms = List.tabulate(arity, {n: Int => new Symbol("x" + n, LocalKind)})
    Lambda(pSyms, Prim(sym, pSyms map Ident))
  }

  private def copyPos(srcTree: Tree)(dstTree: Tree) = {
    dstTree.pos = srcTree.pos
    dstTree
  }
}
