package minischeme

import java.io.PrintWriter
import scala.text.Document
import scala.text.Document._

import SymbolicTreeModule._

/**
 * Format the tree on 80 columns.
 *
 * @author Michel Schinz
 */
object TreePrinter {
  def print(trees: List[Tree]): Unit =
    trees foreach print

  def print(tree: Tree): Unit = {
    val doc = toDoc(tree)
    val writer = new PrintWriter(System.out)
    doc.format(80, writer)
    writer.println()
    writer.flush()
  }

  private def toDoc(tree: Tree): Document = tree match {
    case Define(name, value) =>
      nest(2, group("(define " :: toDoc(name) :/: toDoc(value) :: ")" :: empty))
    case Lambda(ns, body) =>
      nest(2, group("(lambda "
                    :: nest(7, group("(" :: fold(ns map toDoc) :: ")" :: empty))
                    :/: toDoc(body) :: ")" :: empty))
    case Begin(bodies) =>
      nest(2, group("(begin" :/: toDoc(bodies) :: ")" :: empty))
    case Let(name, value, body) =>
      nest(2, group("(let "
                    :: nest(6 + name.toString.length, "((" :: toDoc(name)
                            :: " " :: toDoc(value) :: "))" :: empty)
                    :/: toDoc(body) :: ")" :: empty))
    case If(cond, thenp, elsep) =>
      nest(2, group("(if " :: nest(2, toDoc(cond))
                    :/: toDoc(thenp)
                    :/: toDoc(elsep) :: ")" :: empty))
    case And(c1, c2) =>
      nest(5, group("(and " :: toDoc(c1) :/: toDoc(c2) :: empty))
    case Or(c1, c2) =>
      nest(4, group("(or " :: toDoc(c1) :/: toDoc(c2) :: empty))
    case Apply(fun, args) =>
      nest(1, group("(" :: toDoc(fun :: args) :: ")" :: empty))
    case Prim(sym, args) =>
      nest(1, group("(" :: sym.name :/: group(toDoc(args)) :: ")" :: empty))
    case Ident(name) =>
      toDoc(name)
    case Num(value) =>
      text(String.valueOf(value))
  }

  private def toDoc(symbol: Symbol): Document =
    text(symbol.toString)

  private def toDoc(trees: List[Tree]): Document =
    fold(trees map toDoc)

  private def fold(documents: List[Document]): Document = documents match {
    case Nil => empty
    case _ => documents reduceLeft { _ :/: _ }
  }
}
