#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include "opcode.h"
#include "fail.h"
#include "memory.h"
#include "engine.h"

char* engine_identity = "switch-based";

typedef struct {
  opcode_t opcode;
  opcode_t next[0];
} instr_t;

typedef struct {
  opcode_t opcode;
  reg_idx_t r;
  opcode_t next[0];
} instr_r_t;

typedef struct {
  opcode_t opcode;
  reg_idx_t ra, rb;
  opcode_t next[0];
} instr_rr_t;

typedef struct {
  opcode_t opcode;
  reg_idx_t r;
  intptr_t c;
  opcode_t next[0];
} instr_rc_t;

typedef struct {
  opcode_t opcode;
  reg_idx_t ra, rb, rc;
  opcode_t next[0];
} instr_rrr_t;

static void* curr_addr;
static intptr_t regs[REGS_COUNT];

static void check_available_memory(size_t size);

void engine_setup() {
  curr_addr = memory_get_start();

  for (unsigned r = 0; r < REGS_COUNT; ++r)
    regs[r] = 0;
}

void engine_cleanup() {
  // nothing to do
}

void engine_emit(opcode_t opcode) {
  check_available_memory(sizeof(instr_t));
  instr_t* instr = curr_addr;
  instr->opcode = opcode;
  curr_addr = instr->next;
}

void engine_emit_r(opcode_t opcode, reg_idx_t r) {
  check_available_memory(sizeof(instr_r_t));
  instr_r_t* instr = curr_addr;
  instr->opcode = opcode;
  instr->r = r;
  curr_addr = instr->next;
}

void engine_emit_rr(opcode_t opcode, reg_idx_t ra, reg_idx_t rb) {
  check_available_memory(sizeof(instr_rr_t));
  instr_rr_t* instr = curr_addr;
  instr->opcode = opcode;
  instr->ra = ra;
  instr->rb = rb;
  curr_addr = instr->next;
}

void engine_emit_rc(opcode_t opcode, reg_idx_t r, intptr_t c) {
  check_available_memory(sizeof(instr_rc_t));
  instr_rc_t* instr = curr_addr;
  instr->opcode = opcode;
  instr->r = r;
  instr->c = c;
  curr_addr = instr->next;
}

void engine_emit_rrr(opcode_t opcode,
                     reg_idx_t ra,
                     reg_idx_t rb,
                     reg_idx_t rc) {
  check_available_memory(sizeof(instr_rrr_t));
  instr_rrr_t* instr = curr_addr;
  instr->opcode = opcode;
  instr->ra = ra;
  instr->rb = rb;
  instr->rc = rc;
  curr_addr = instr->next;
}

void* engine_get_next_address() {
  return curr_addr;
}

void engine_patch(void* instr_to_patch, void* target) {
  instr_rc_t* instr = instr_to_patch;
  instr->c = (intptr_t)target;
}

int engine_run() {
  opcode_t* pc = memory_get_start();
  for (;;) {
    switch (*pc) {
    case opcode_ADD: {
      instr_rrr_t* instr = (instr_rrr_t*)pc;
      regs[instr->ra] = regs[instr->rb] + regs[instr->rc];
      pc = instr->next;
    } break;

    case opcode_SUB: {
      instr_rrr_t* instr = (instr_rrr_t*)pc;
      regs[instr->ra] = regs[instr->rb] - regs[instr->rc];
      pc = instr->next;
    } break;

    case opcode_MUL: {
      instr_rrr_t* instr = (instr_rrr_t*)pc;
      regs[instr->ra] = regs[instr->rb] * regs[instr->rc];
      pc = instr->next;
    } break;

    case opcode_DIV: {
      instr_rrr_t* instr = (instr_rrr_t*)pc;
      if (regs[instr->rc] == 0)
        fail("division by zero");
      regs[instr->ra] = regs[instr->rb] / regs[instr->rc];
      pc = instr->next;
    } break;

    case opcode_MOD: {
      instr_rrr_t* instr = (instr_rrr_t*)pc;
      if (regs[instr->rc] == 0)
        fail("division by zero");
      regs[instr->ra] = regs[instr->rb] % regs[instr->rc];
      pc = instr->next;
    } break;

    case opcode_JLT: {
      instr_rrr_t* instr = (instr_rrr_t*)pc;
      pc = (regs[instr->rb] < regs[instr->rc])
        ? (opcode_t*)regs[instr->ra]
        : instr->next;
    } break;

    case opcode_JLE: {
      instr_rrr_t* instr = (instr_rrr_t*)pc;
      pc = (regs[instr->rb] <= regs[instr->rc])
        ? (opcode_t*)regs[instr->ra]
        : instr->next;
    } break;

    case opcode_JEQ: {
      instr_rrr_t* instr = (instr_rrr_t*)pc;
      pc = (regs[instr->rb] == regs[instr->rc])
        ? (opcode_t*)regs[instr->ra]
        : instr->next;
    } break;

    case opcode_JNE: {
      instr_rrr_t* instr = (instr_rrr_t*)pc;
      pc = (regs[instr->rb] != regs[instr->rc])
        ? (opcode_t*)regs[instr->ra]
        : instr->next;
    } break;

    case opcode_JGE: {
      instr_rrr_t* instr = (instr_rrr_t*)pc;
      pc = (regs[instr->rb] >= regs[instr->rc])
        ? (opcode_t*)regs[instr->ra]
        : instr->next;
    } break;

    case opcode_JGT: {
      instr_rrr_t* instr = (instr_rrr_t*)pc;
      pc = (regs[instr->rb] > regs[instr->rc])
        ? (opcode_t*)regs[instr->ra]
        : instr->next;
    } break;

    case opcode_LINT: {
      instr_rc_t* instr = (instr_rc_t*)pc;
      regs[instr->r] = instr->c;
      pc = instr->next;
    } break;

    case opcode_LOAD: {
      instr_rrr_t* instr = (instr_rrr_t*)pc;
      regs[instr->ra] = ((intptr_t*)regs[instr->rb])[(int)regs[instr->rc]];
      pc = instr->next;
    } break;

    case opcode_STOR: {
      instr_rrr_t* instr = (instr_rrr_t*)pc;
      ((intptr_t*)regs[instr->rb])[(int)regs[instr->rc]] = regs[instr->ra];
      pc = instr->next;
    } break;

    case opcode_ALOC: {
      instr_rr_t* instr = (instr_rr_t*)pc;
      regs[instr->ra] = (intptr_t)memory_allocate(regs[instr->rb]);
      pc = instr->next;
    } break;

    case opcode_RCHR: {
      instr_r_t* instr = (instr_r_t*)pc;
      char input;
      fscanf(stdin, "%c", &input);
      regs[instr->r] = (intptr_t)input;
      pc = instr->next;
    } break;

    case opcode_PCHR: {
      instr_r_t* instr = (instr_r_t*)pc;
      fprintf(stdout, "%c", (char)regs[instr->r]);
      fflush(stdout);
      pc = instr->next;
    } break;

    case opcode_HALT:
      return 0;
    }
  }
}

intptr_t engine_get_register(reg_idx_t r) {
  return regs[r];
}

static void check_available_memory(size_t size) {
  if (curr_addr + size > memory_get_end())
    fail("not enough memory for code");
}
